from dataclasses import dataclass

import pandas as pd

from order_matching.execution import Execution
from order_matching.side import Side


@dataclass(kw_only=True)
class Trade:
    """Single trade storage class."""

    side: Side
    price: float
    size: float
    incoming_order_id: str
    book_order_id: str
    execution: Execution
    trade_id: str
    timestamp: pd.Timestamp = None



import pandas as pd

from order_matching.order import LimitOrder
from order_matching.order_book import OrderBook
from order_matching.orders import Orders
from order_matching.schemas import OrderBookSummarySchema
from order_matching.side import Side


class TestOrderBook:
    timestamp = pd.Timestamp.now()

    def test_init(self) -> None:
        order_book = OrderBook()

        assert order_book.bids == {}
        assert order_book.offers == {}
        assert order_book.current_price == float("inf")

    def test_append_with_same_price(self) -> None:
        order_book = OrderBook()

        buy_price, sell_price = 1.2, 1.3
        buy_orders = [
            LimitOrder(
                side=Side.BUY, price=buy_price, size=size, timestamp=self.timestamp, order_id=order_id, trader_id="x"
            )
            for size, order_id in zip([12, 65, 98], ["a", "b", "c"], strict=True)
        ]
        sell_orders = [
            LimitOrder(
                side=Side.SELL, price=sell_price, size=size, timestamp=self.timestamp, order_id=order_id, trader_id="y"
            )
            for size, order_id in zip([12, 65, 98], ["a", "b", "c"], strict=True)
        ]
        for order in [*buy_orders, *sell_orders]:
            order_book.append(incoming_order=order)

        assert order_book.bids == {buy_price: Orders(buy_orders)}
        assert order_book.offers == {sell_price: Orders(sell_orders)}
        assert order_book.current_price == (buy_price + sell_price) / 2

    def test_append(self) -> None:
        order_book = OrderBook()

        buy_order_first = LimitOrder(
            side=Side.BUY, price=1.2, size=2.3, timestamp=self.timestamp, order_id="xyz", trader_id="x"
        )
        order_book.append(incoming_order=buy_order_first)

        assert order_book.bids == {buy_order_first.price: Orders([buy_order_first])}
        assert order_book.offers == dict()
        assert order_book.current_price == float("inf")

        sell_order_first = LimitOrder(
            side=Side.SELL, price=3.4, size=5.6, timestamp=self.timestamp, order_id="abc", trader_id="x"
        )
        order_book.append(incoming_order=sell_order_first)

        assert order_book.bids == {buy_order_first.price: Orders([buy_order_first])}
        assert order_book.offers == {sell_order_first.price: Orders([sell_order_first])}
        assert order_book.current_price == 2.3

        buy_order_second = LimitOrder(
            side=Side.BUY,
            price=buy_order_first.price,
            size=6.7,
            timestamp=self.timestamp,
            order_id="qwe",
            trader_id="x",
        )
        sell_order_second = LimitOrder(
            side=Side.SELL, price=5.9, size=9.3, timestamp=self.timestamp, order_id="abc", trader_id="x"
        )
        order_book.append(incoming_order=buy_order_second)
        order_book.append(incoming_order=sell_order_second)

        assert order_book.bids == {buy_order_first.price: Orders([buy_order_first, buy_order_second])}
        assert order_book.offers == {
            sell_order_first.price: Orders([sell_order_first]),
            sell_order_second.price: Orders([sell_order_second]),
        }
        assert order_book.current_price == 2.3

    def test_append_remove_and_get_subset(self) -> None:
        order_book = OrderBook()

        assert order_book.get_subset(expiration=self.timestamp) == Orders()

        first_buy_order = self._get_sample_orders().orders[0]
        second_buy_order = self._get_sample_orders().orders[2]
        third_buy_order = self._get_sample_orders().orders[4]
        timedelta = pd.Timedelta(1, unit="D")
        third_buy_order.expiration = self.timestamp + timedelta

        assert first_buy_order.price == second_buy_order.price
        assert first_buy_order.price != third_buy_order.price
        assert first_buy_order.side == second_buy_order.side
        assert first_buy_order.side == third_buy_order.side
        assert first_buy_order.expiration == second_buy_order.expiration
        assert first_buy_order.expiration != third_buy_order.expiration
        assert order_book.bids == dict()

        order_book.remove(incoming_order=first_buy_order)

        assert order_book.bids == dict()
        assert order_book.get_subset(expiration=self.timestamp) == Orders()

        order_book.append(incoming_order=first_buy_order)
        order_book.append(incoming_order=second_buy_order)
        order_book.append(incoming_order=third_buy_order)

        assert order_book.bids == {
            first_buy_order.price: Orders([first_buy_order, second_buy_order]),
            third_buy_order.price: Orders([third_buy_order]),
        }
        assert order_book.get_subset(expiration=self.timestamp) == Orders([first_buy_order, second_buy_order])
        assert order_book.get_subset(expiration=self.timestamp + timedelta) == Orders([third_buy_order])

        order_book.remove(incoming_order=first_buy_order)

        assert order_book.bids == {
            first_buy_order.price: Orders([second_buy_order]),
            third_buy_order.price: Orders([third_buy_order]),
        }
        assert order_book.get_subset(expiration=self.timestamp) == Orders([second_buy_order])
        assert order_book.get_subset(expiration=self.timestamp + timedelta) == Orders([third_buy_order])

        order_book.remove(incoming_order=second_buy_order)

        assert order_book.bids == {third_buy_order.price: Orders([third_buy_order])}
        assert order_book.get_subset(expiration=self.timestamp) == Orders()
        assert order_book.get_subset(expiration=self.timestamp + timedelta) == Orders([third_buy_order])

        order_book.remove(incoming_order=third_buy_order)

        assert order_book.bids == dict()
        assert order_book.get_subset(expiration=self.timestamp) == Orders()

    def test_order_book_summary(self) -> None:
        order_book = OrderBook()
        for order in self._get_sample_orders():
            order_book.append(incoming_order=order)
        summary = order_book.summary()

        OrderBookSummarySchema.validate(summary, lazy=True)
        assert summary.index.is_monotonic_increasing
        assert summary[OrderBookSummarySchema.price].is_monotonic_increasing
        assert order_book.current_price == 2.3

    def test_order_book_imbalance_one_buy_order(self) -> None:
        order_book = OrderBook()
        buy_orders = [
            LimitOrder(side=Side.BUY, price=1.3, size=38, timestamp=self.timestamp, order_id="x", trader_id="x")
        ]
        for order in [*buy_orders]:
            order_book.append(incoming_order=order)

        assert order_book.get_imbalance(price_range=0.1) == 1

    def test_order_book_imbalance_one_sell_order(self) -> None:
        order_book = OrderBook()
        sell_orders = [
            LimitOrder(side=Side.SELL, price=1.3, size=38, timestamp=self.timestamp, order_id="x", trader_id="x")
        ]
        for order in [*sell_orders]:
            order_book.append(incoming_order=order)

        assert order_book.get_imbalance(price_range=0.1) == -1

    def test_order_book_imbalance_with_equal_orders(self) -> None:
        order_book = OrderBook()
        buy_orders = [
            LimitOrder(side=Side.BUY, price=1.4, size=10, timestamp=self.timestamp, order_id="x", trader_id="x")
        ]
        sell_orders = [
            LimitOrder(side=Side.SELL, price=1.6, size=10, timestamp=self.timestamp, order_id="x", trader_id="y"),
            LimitOrder(side=Side.SELL, price=1.8, size=20, timestamp=self.timestamp, order_id="x", trader_id="y"),
        ]
        for order in [*buy_orders, *sell_orders]:
            order_book.append(incoming_order=order)

        assert order_book.get_imbalance(price_range=0.05) == 0
        assert order_book.get_imbalance(price_range=0.1) == 0
        assert order_book.get_imbalance(price_range=0.2) == 0
        assert order_book.get_imbalance(price_range=0.3) == (10 - 20 - 10) / (10 + 20 + 10)

    def test_order_book_imbalance_with_several_orders(self) -> None:
        order_book = OrderBook()
        buy_orders = [
            LimitOrder(side=Side.BUY, price=price, size=size, timestamp=self.timestamp, order_id="x", trader_id="x")
            for size, price in zip([12, 65, 98], [1.1, 1.3, 1.4], strict=True)
        ]
        sell_orders = [
            LimitOrder(side=Side.SELL, price=price, size=size, timestamp=self.timestamp, order_id="x", trader_id="y")
            for size, price in zip([8, 86, 72], [1.5, 1.7, 1.8], strict=True)
        ]
        for order in [*buy_orders, *sell_orders]:
            order_book.append(incoming_order=order)

        assert order_book.get_imbalance(price_range=0.1) == (98 - 8) / (98 + 8)
        assert order_book.get_imbalance(price_range=0.2) == (65 + 98 - 8) / (65 + 98 + 8)
        assert order_book.get_imbalance(price_range=0.3) == (65 + 98 - 8 - 86) / (65 + 98 + 8 + 86)
        assert order_book.get_imbalance(price_range=0.4) == (12 + 65 + 98 - 8 - 86 - 72) / (12 + 65 + 98 + 8 + 86 + 72)

    def _get_sample_orders(self) -> Orders:
        orders = [
            LimitOrder(side=Side.BUY, price=1.2, size=2.3, timestamp=self.timestamp, order_id="xyz", trader_id="x"),
            LimitOrder(side=Side.SELL, price=3.4, size=5.6, timestamp=self.timestamp, order_id="abc", trader_id="x"),
            LimitOrder(side=Side.BUY, price=1.2, size=6.7, timestamp=self.timestamp, order_id="qwe", trader_id="x"),
            LimitOrder(side=Side.SELL, price=5.9, size=9.3, timestamp=self.timestamp, order_id="abc", trader_id="x"),
            LimitOrder(side=Side.BUY, price=1.1, size=6.7, timestamp=self.timestamp, order_id="qwe", trader_id="x"),
        ]
        for order in orders:
            order.expiration = self.timestamp
        return Orders(orders)



import pandas as pd
import pytest

from order_matching.execution import Execution
from order_matching.order import LimitOrder, MarketOrder, Order
from order_matching.side import Side
from order_matching.status import Status


class TestOrder:
    @pytest.mark.parametrize("side", [Side.BUY, Side.SELL])
    @pytest.mark.parametrize("price", [1.2, 2.4])
    @pytest.mark.parametrize("size", [10, 4.1])
    @pytest.mark.parametrize("timestamp", pd.date_range(start="2022", periods=3))
    @pytest.mark.parametrize("expiration", [None, *pd.date_range(start="2023", periods=3)])
    @pytest.mark.parametrize("order_id", ["a", "b"])
    @pytest.mark.parametrize("trader_id", ["x", "y"])
    @pytest.mark.parametrize("execution", [Execution.LIMIT, Execution.MARKET])
    @pytest.mark.parametrize("status", [Status.OPEN, Status.CANCEL])
    @pytest.mark.parametrize("price_number_of_digits", [1, 3])
    def test_order_required_defaults(
        self,
        side: Side,
        price: float,
        size: float,
        timestamp: pd.Timestamp,
        expiration: pd.Timestamp,
        order_id: str,
        trader_id: str,
        execution: Execution,
        status: Status,
        price_number_of_digits: int,
    ) -> None:
        order = Order(
            side=side,
            price=price,
            size=size,
            timestamp=timestamp,
            expiration=expiration,
            order_id=order_id,
            trader_id=trader_id,
            execution=execution,
            status=status,
            price_number_of_digits=price_number_of_digits,
        )
        assert order.side == side
        assert order.price == round(number=price, ndigits=price_number_of_digits)
        assert order.size == size
        assert order.timestamp == timestamp
        assert order.expiration == expiration
        assert order.order_id == order_id
        assert order.trader_id == trader_id
        assert order.execution == execution
        assert order.status == status
        assert order.price_number_of_digits == price_number_of_digits


class TestLimitOrder:
    @pytest.mark.parametrize("side", [Side.BUY, Side.SELL])
    @pytest.mark.parametrize("price", [1.2, 2.4])
    @pytest.mark.parametrize("size", [10, 4.1])
    @pytest.mark.parametrize("timestamp", pd.date_range(start="2022", periods=3))
    @pytest.mark.parametrize("order_id", ["a", "b"])
    @pytest.mark.parametrize("trader_id", ["x", "y"])
    @pytest.mark.parametrize("price_number_of_digits", [1, 3])
    def test_order_required_defaults(
        self,
        side: Side,
        price: float,
        size: float,
        timestamp: pd.Timestamp,
        order_id: str,
        trader_id: str,
        price_number_of_digits: int,
    ) -> None:
        order = LimitOrder(
            side=side,
            price=price,
            size=size,
            timestamp=timestamp,
            order_id=order_id,
            trader_id=trader_id,
            price_number_of_digits=price_number_of_digits,
        )
        assert order.side == side
        assert order.price == round(number=price, ndigits=price_number_of_digits)
        assert order.size == size
        assert order.timestamp == timestamp
        assert order.order_id == order_id
        assert order.trader_id == trader_id
        assert order.execution == Execution.LIMIT
        assert order.status == Status.OPEN
        assert order.price_number_of_digits == price_number_of_digits


class TestMarketOrder:
    @pytest.mark.parametrize("side", [Side.BUY, Side.SELL])
    @pytest.mark.parametrize("size", [10, 4.1])
    @pytest.mark.parametrize("timestamp", pd.date_range(start="2022", periods=3))
    @pytest.mark.parametrize("order_id", ["a", "b"])
    @pytest.mark.parametrize("trader_id", ["x", "y"])
    def test_order_required_defaults(
        self, side: Side, size: float, timestamp: pd.Timestamp, order_id: str, trader_id: str
    ) -> None:
        order = MarketOrder(side=side, size=size, timestamp=timestamp, order_id=order_id, trader_id=trader_id)
        assert order.side == side
        assert order.price == 0 if side == side.SELL else float("inf")
        assert order.size == size
        assert order.timestamp == timestamp
        assert order.order_id == order_id
        assert order.trader_id == trader_id
        assert order.execution == Execution.MARKET
        assert order.status == Status.OPEN
        assert order.price_number_of_digits == 1



from __future__ import annotations

from collections import defaultdict
from dataclasses import asdict

import pandas as pd
from pandera.typing import DataFrame

from order_matching.schemas import TradeDataSchema
from order_matching.trade import Trade


class ExecutedTrades:
    """Executed Trades.

    Storage class for collections of trades.

    Parameters
    ----------
    trades
        Trades
    """

    def __init__(self, trades: list[Trade] = None) -> None:
        self._trades: dict[pd.Timestamp, list[Trade]] = defaultdict(list)
        if trades:
            self.add(trades=trades)

    @property
    def trades(self) -> list[Trade]:
        """List of trades."""
        trades = list()
        for same_time_trades in self._trades.values():
            trades.extend(same_time_trades)
        return trades

    def add(self, trades: list[Trade]) -> None:
        """Add new trades to the object.

        Parameters
        ----------
        trades
            List of new trades to append to existing ones
        """
        for trade in trades:
            self._trades[trade.timestamp].append(trade)

    def get(self, timestamp: pd.Timestamp) -> list[Trade]:
        """Get subset by timestamp.

        Returns
        -------
        list[Trade]
            List of trades with the same timestamp
        """
        return self._trades[timestamp]

    def to_frame(self) -> DataFrame[TradeDataSchema]:
        """Get pandas DataFrame of all stored trades.

        Returns
        -------
        DataFrame[TradeDataSchema]
            pandas DataFrame of all stored trades
        """
        trades = self.trades
        if len(trades) == 0:
            return pd.DataFrame()
        else:
            return pd.DataFrame.from_records([asdict(trade) for trade in trades]).assign(
                **{
                    TradeDataSchema.side: lambda df: df[TradeDataSchema.side].astype(str),
                    TradeDataSchema.execution: lambda df: df[TradeDataSchema.execution].astype(str),
                }
            )

    def __add__(self, other: ExecutedTrades) -> ExecutedTrades:
        trades = ExecutedTrades()
        trades.add(trades=self.trades)
        trades.add(trades=other.trades)
        return trades



import pandas as pd
import pytest

from order_matching.order import LimitOrder
from order_matching.orders import Orders
from order_matching.random import get_faker, get_random_generator
from order_matching.side import Side


@pytest.fixture
def random_orders() -> Orders:
    faker = get_faker()
    rng = get_random_generator(seed=42)
    orders_per_timestamp, number_of_time_points = 100, 10
    time_intervals = rng.uniform(low=0, high=1, size=number_of_time_points)
    random_timestamps = pd.Timestamp(2023, 1, 1) + pd.to_timedelta(time_intervals.cumsum(), unit="D")
    orders = list()
    for timestamp in random_timestamps:
        prices = rng.lognormal(mean=1, size=orders_per_timestamp).round(decimals=1)
        sizes = rng.lognormal(mean=1, size=orders_per_timestamp).round(decimals=4)
        sides = rng.choice([Side.SELL, Side.BUY], size=orders_per_timestamp)
        new_orders = [
            LimitOrder(
                side=side, price=price, size=size, timestamp=timestamp, order_id=faker.uuid4(), trader_id=faker.uuid4()
            )
            for side, price, size in zip(sides, prices, sizes, strict=True)
        ]
        orders.extend(new_orders)
    return Orders(orders)



from enum import Enum


class CustomEnum(Enum):
    """Custom enumerator.

    Used across the library as a base class for enumeration objects.
    """

    def __lt__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self.value < other.value
        else:
            return NotImplemented

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name



from pandera import Field, SchemaModel
from pandera.typing import DateTime, Series

from order_matching.execution import Execution
from order_matching.side import Side
from order_matching.status import Status


class BaseOrderSchema(SchemaModel):
    """Base order schema."""

    side: Series[str] = Field(isin=[Side.BUY.name, Side.SELL.name])
    price: Series[float] = Field(gt=0)
    size: Series[float] = Field(gt=0)


class OrderBookSummarySchema(BaseOrderSchema):
    """Order book summary schema."""

    price: Series[float] = Field(unique=True, gt=0)
    count: Series[int] = Field(ge=0)

    class Config:
        strict = True


class OrderDataSchema(BaseOrderSchema):
    """Order data schema."""

    timestamp: Series[DateTime]
    expiration: Series[DateTime] = Field(nullable=True)
    order_id: Series[str]
    trader_id: Series[str]
    execution: Series[str] = Field(isin=[Execution.MARKET.name, Execution.LIMIT.name])
    status: Series[str] = Field(isin=[Status.OPEN.name, Status.CANCEL.name])
    price_number_of_digits: Series[int]

    class Config:
        strict = True


class TradeDataSchema(BaseOrderSchema):
    """Trade data schema."""

    timestamp: Series[DateTime]
    incoming_order_id: Series[str]
    book_order_id: Series[str]
    trade_id: Series[str]
    execution: Series[str] = Field(isin=[Execution.MARKET.name, Execution.LIMIT.name])

    class Config:
        strict = True



from collections import defaultdict
from typing import cast

import pandas as pd
from pandera.typing import DataFrame

from order_matching.order import Order
from order_matching.orders import Orders
from order_matching.schemas import OrderBookSummarySchema
from order_matching.side import Side

OrderBookOrdersType = dict[float, Orders]


class OrderBook:
    """Order Book storage class."""

    def __init__(self) -> None:
        self.bids: OrderBookOrdersType = defaultdict(Orders)
        self.offers: OrderBookOrdersType = defaultdict(Orders)
        self.orders_by_expiration: dict[pd.Timestamp, Orders] = defaultdict(Orders)

    def append(self, incoming_order: Order) -> None:
        """Add one order to the order book.

        Parameters
        ----------
        incoming_order
            New order
        """
        orders = self._get_same_side_orders(incoming_order=incoming_order)
        orders[incoming_order.price].add(orders=[incoming_order])
        self.orders_by_expiration[incoming_order.expiration].add(orders=[incoming_order])

    def remove(self, incoming_order: Order) -> None:
        """Remove one order from the order book.

        Parameters
        ----------
        incoming_order
            Order to be removed
        """
        orders = self._get_same_side_orders(incoming_order=incoming_order)
        orders[incoming_order.price].remove(orders=[incoming_order])
        self.orders_by_expiration[incoming_order.expiration].remove(orders=[incoming_order])
        if len(orders[incoming_order.price]) == 0:
            orders.pop(incoming_order.price)
        if len(self.orders_by_expiration[incoming_order.expiration]) == 0:
            self.orders_by_expiration.pop(incoming_order.expiration)

    def summary(self) -> DataFrame[OrderBookSummarySchema]:
        """Summary of the order book as a pandas DataFrame.

        Returns
        -------
        DataFrame[OrderBookSummarySchema]
            Summary of the order book as a pandas DataFrame
        """
        bids = pd.DataFrame(
            {
                OrderBookSummarySchema.side: Side.BUY.name,
                OrderBookSummarySchema.price: self._get_bid_prices(),
                OrderBookSummarySchema.size: self._get_bid_sizes(),
                OrderBookSummarySchema.count: self._get_bid_counts(),
            }
        )
        offers = pd.DataFrame(
            {
                OrderBookSummarySchema.side: Side.SELL.name,
                OrderBookSummarySchema.price: self._get_offer_prices(),
                OrderBookSummarySchema.size: self._get_offer_sizes(),
                OrderBookSummarySchema.count: self._get_offer_counts(),
            }
        )
        return pd.concat([bids, offers], ignore_index=True).assign(
            **{OrderBookSummarySchema.count: lambda df: df[OrderBookSummarySchema.count].astype(int)}
        )

    @property
    def current_price(self) -> float:
        """Current market price."""
        return (self.max_bid + self.min_offer) / 2

    def get_opposite_side_orders(self, incoming_order: Order) -> OrderBookOrdersType:
        """Get orders on the opposite side.

        If incoming order is sell, then bids are returned.
        If incoming order is buy, then offers are returned.

        Parameters
        ----------
        incoming_order

        Returns
        -------
        OrderBookOrdersType
        """
        match incoming_order.side:  # noqa E501
            case Side.SELL:
                return self.bids
            case Side.BUY:
                return self.offers

    def get_subset(self, expiration: pd.Timestamp) -> Orders:
        """Get orders with given expiration time.

        Parameters
        ----------
        expiration

        Returns
        -------
        Orders
        """
        return self.orders_by_expiration[expiration]

    def matching_order_exists(self, incoming_order: Order) -> bool:
        """Check that matching order exists.

        Parameters
        ----------
        incoming_order

        Returns
        -------
        bool
        """
        match incoming_order.side:
            case Side.SELL:
                return incoming_order.price <= self.max_bid and len(self.bids) > 0
            case Side.BUY:
                return incoming_order.price >= self.min_offer and len(self.offers) > 0

    def get_matching_sorted_opposite_side_prices(self, incoming_order: Order) -> list[float]:
        """Get sorted prices of the matching orders on the opposite side.

        Parameters
        ----------
        incoming_order

        Returns
        -------
        list[float]
        """
        prices = self._get_sorted_opposite_side_prices(incoming_order=incoming_order)
        match incoming_order.side:
            case Side.SELL:
                return list(filter(lambda price: price >= incoming_order.price, prices))
            case Side.BUY:
                return list(filter(lambda price: price <= incoming_order.price, prices))

    def get_imbalance(self, price_range: float = 0.1) -> float:
        r"""Calculate order book imbalance.

        Order book imbalance indicator is defined as
        $$
        \rho(L)=\frac{V_L^b-V_L^a}{V_L^b+V_L^a}\in[-1,1],
        $$
        where
        $$
        V_L^b=\sum_i Q_i^b\cdot\mathbf{1}\left[P_i^b\geq P^M-L\right],
        $$
        and
        $$
        V_L^a=\sum_i Q_i^a\cdot\mathbf{1}\left[P_i^a\leq P^M+L\right],
        $$
        are the total bid/ask volumes within $L>0$ from the market price.
        It is said that the market is balanced when $\rho\approx0$.
        The market is imbalanced and there exists demand (supply) pressure when $\rho>0$ ($\rho<0$).
        As it is clear from the formula,
        the imbalance can be computed at any distance $L$ from the market price.
        The higher the distance, the deeper looks this indicator into the order book.

        Parameters
        ----------
        price_range
            Left/right from the current market price

        Returns
        -------
        float
            Market imbalance indicator
        """
        summary = self.summary()
        if summary.empty:
            return 0
        elif summary[summary[OrderBookSummarySchema.side] == Side.SELL.name].empty:
            return 1
        elif summary[summary[OrderBookSummarySchema.side] == Side.BUY.name].empty:
            return -1
        else:
            return self._get_non_trivial_imbalance(price_range=price_range)

    def _get_non_trivial_imbalance(self, price_range: float) -> float:
        schema = OrderBookSummarySchema
        upper_bound = self.current_price + price_range
        lower_bound = self.current_price - price_range
        summary_subset = self.summary().pipe(lambda df: df[df[schema.price].between(lower_bound, upper_bound)])
        summary_subset = cast(pd.DataFrame, summary_subset)
        buy_volume = summary_subset.loc[summary_subset[schema.side] == Side.BUY.name, schema.size].sum()
        sell_volume = summary_subset.loc[summary_subset[schema.side] == Side.SELL.name, schema.size].sum()
        if buy_volume + sell_volume > 0:
            return (buy_volume - sell_volume) / (buy_volume + sell_volume)
        else:
            return 0

    def _get_same_side_orders(self, incoming_order: Order) -> OrderBookOrdersType:
        match incoming_order.side:
            case Side.SELL:
                return self.offers
            case Side.BUY:
                return self.bids

    def _get_sorted_opposite_side_prices(self, incoming_order: Order) -> list[float]:
        is_sell_side = incoming_order.side == Side.SELL
        return sorted(self.get_opposite_side_orders(incoming_order=incoming_order).keys(), reverse=is_sell_side)

    def _get_bid_prices(self) -> list[float]:
        return self._get_order_prices(orders=self.bids)

    def _get_offer_prices(self) -> list[float]:
        return self._get_order_prices(orders=self.offers)

    def _get_bid_sizes(self) -> list[float]:
        return self._get_order_sizes(orders=self.bids, prices=self._get_bid_prices())

    def _get_offer_sizes(self) -> list[float]:
        return self._get_order_sizes(orders=self.offers, prices=self._get_offer_prices())

    def _get_bid_counts(self) -> list[int]:
        return self._get_order_counts(orders=self.bids, prices=self._get_bid_prices())

    def _get_offer_counts(self) -> list[int]:
        return self._get_order_counts(orders=self.offers, prices=self._get_offer_prices())

    @staticmethod
    def _get_order_prices(orders: OrderBookOrdersType) -> list[float]:
        return sorted(orders.keys())

    @staticmethod
    def _get_order_sizes(orders: OrderBookOrdersType, prices: list[float]) -> list[float]:
        return [sum(order.size for order in orders[price]) for price in prices]

    @staticmethod
    def _get_order_counts(orders: OrderBookOrdersType, prices: list[float]) -> list[int]:
        return [len(orders[price]) for price in prices]

    @property
    def max_bid(self) -> float:
        """Maximum bid price."""
        if self.bids:
            return max(self.bids.keys())
        else:
            return 0.0

    @property
    def min_offer(self) -> float:
        """Minimum offer price."""
        if self.offers:
            return min(self.offers.keys())
        else:
            return float("inf")



from dataclasses import dataclass, field

import pandas as pd

from order_matching.execution import Execution
from order_matching.side import Side
from order_matching.status import Status


@dataclass(kw_only=True)
class Order:
    """Single order base storage class."""

    side: Side
    price: float
    size: float
    timestamp: pd.Timestamp
    order_id: str
    trader_id: str
    execution: Execution
    expiration: pd.Timestamp = pd.NaT
    status: Status = Status.OPEN
    price_number_of_digits: int = 1

    def __post_init__(self) -> None:
        self.price = round(number=self.price, ndigits=self.price_number_of_digits)


@dataclass(kw_only=True)
class LimitOrder(Order):
    """Single limit order storage class."""

    execution: Execution = field(init=False, default=Execution.LIMIT)


@dataclass(kw_only=True)
class MarketOrder(Order):
    """Single market order storage class."""

    execution: Execution = field(init=False, default=Execution.MARKET)
    price: float = field(init=False)

    def __post_init__(self) -> None:
        self.price = 0 if self.side == Side.SELL else float("inf")



from dataclasses import asdict

import pandas as pd
import pytest

from order_matching.execution import Execution
from order_matching.schemas import TradeDataSchema
from order_matching.side import Side
from order_matching.trade import Trade


class TestTrade:
    @pytest.mark.parametrize("side", [Side.BUY, Side.SELL])
    @pytest.mark.parametrize("price", [1.2, 2.4])
    @pytest.mark.parametrize("size", [10.0, 4.1])
    @pytest.mark.parametrize("timestamp", pd.date_range(start="2022", periods=3))
    @pytest.mark.parametrize("incoming_order_id", ["a", "abc"])
    @pytest.mark.parametrize("book_order_id", ["a", "abc"])
    @pytest.mark.parametrize("trade_id", ["t", "tr"])
    @pytest.mark.parametrize("execution", [Execution.LIMIT, Execution.MARKET])
    def test_trade_required_defaults(
        self,
        side: Side,
        price: float,
        size: float,
        timestamp: pd.Timestamp,
        incoming_order_id: str,
        book_order_id: str,
        trade_id: str,
        execution: Execution,
    ) -> None:
        trade = Trade(
            side=side,
            price=price,
            size=size,
            timestamp=timestamp,
            incoming_order_id=incoming_order_id,
            book_order_id=book_order_id,
            execution=execution,
            trade_id=trade_id,
        )
        assert trade.side == side
        assert trade.price == price
        assert trade.size == size
        assert trade.timestamp == timestamp
        assert trade.incoming_order_id == incoming_order_id
        assert trade.book_order_id == book_order_id
        assert trade.execution == execution
        assert trade.trade_id == trade_id
        trades = pd.DataFrame(asdict(trade), index=pd.Index([0])).assign(
            **{
                TradeDataSchema.side: lambda df: df[TradeDataSchema.side].astype(str),
                TradeDataSchema.execution: lambda df: df[TradeDataSchema.execution].astype(str),
            }
        )
        TradeDataSchema.validate(trades, lazy=True)



from faker import Faker
from numpy.random import Generator, default_rng


def get_random_generator(seed: int = None) -> Generator:
    """Get numpy random number generator.

    Parameters
    ----------
    seed
        Random seed

    Returns
    -------
    Generator
    """
    return default_rng(seed=seed)


def get_faker(seed: int = None) -> Faker:
    """Get Faker random generator.

    Parameters
    ----------
    seed
        Random seed

    Returns
    -------
    Faker
    """
    faker = Faker()
    faker.seed_instance(seed)
    return faker



from order_matching.random import get_faker, get_random_generator


def test_get_faker() -> None:
    faker = get_faker()
    number_of_ids = 100

    assert len({faker.uuid4() for _ in range(number_of_ids)}) == number_of_ids

    faker1, faker2 = get_faker(seed=42), get_faker(seed=42)

    for _ in range(number_of_ids):
        assert faker1.uuid4() == faker2.uuid4()


def test_get_random_generator() -> None:
    rng = get_random_generator()
    number_of_random_numbers = 100

    assert len({rng.normal() for _ in range(number_of_random_numbers)}) == number_of_random_numbers

    rng1, rng2 = get_random_generator(seed=42), get_random_generator(seed=42)

    for _ in range(number_of_random_numbers):
        assert rng1.normal() == rng2.normal()



from __future__ import annotations

from dataclasses import asdict
from typing import Generator, Iterator, Sequence

import pandas as pd
from pandera.typing import DataFrame

from order_matching.order import Order
from order_matching.schemas import OrderDataSchema


class Orders:
    """Storage class for orders.

    Parameters
    ----------
    orders
    """

    def __init__(self, orders: Sequence[Order] = None) -> None:
        self.orders = list() if orders is None else list(orders)
        self._sort_orders_inplace()

    def add(self, orders: list[Order]) -> None:
        """Add new orders to the storage and sort them by timestamp.

        Parameters
        ----------
        orders
        """
        self.orders.extend(orders)
        self._sort_orders_inplace()

    def dequeue(self) -> Order:
        """Get the earliest order and remove it from the storage.

        Returns
        -------
        Order
            Earliest order
        """
        return self.orders.pop(0)

    def remove(self, orders: list[Order]) -> None:
        """Remove a list of orders from the storage matching by `order_id`.

        Partially executed orders may have their properties changed, hence matching is done only based on the id.

        Parameters
        ----------
        orders
        """
        for order_to_remove in orders:
            if order_to_remove.order_id in self._order_ids:
                self.orders.remove(self._get_order(order_id=order_to_remove.order_id))

    def to_frame(self) -> DataFrame[OrderDataSchema]:
        """Get pandas DataFrame with all orders in the storage.

        Returns
        -------
        DataFrame[OrderDataSchema]
        """
        if len(self.orders) == 0:
            return pd.DataFrame()
        else:
            return pd.DataFrame.from_records([asdict(order) for order in self.orders]).assign(
                **{
                    OrderDataSchema.side: lambda df: df[OrderDataSchema.side].astype(str),
                    OrderDataSchema.execution: lambda df: df[OrderDataSchema.execution].astype(str),
                    OrderDataSchema.status: lambda df: df[OrderDataSchema.status].astype(str),
                }
            )

    @property
    def is_empty(self) -> bool:
        """Check if the storage is empty."""
        return len(self.orders) == 0

    def __add__(self, other: Orders) -> Orders:
        orders = Orders()
        orders.add(orders=self.orders)
        orders.add(orders=other.orders)
        return orders

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Orders):
            return NotImplemented
        else:
            return self.orders == other.orders

    def __iter__(self) -> Iterator[Order]:
        return iter(self.orders)

    def __next__(self) -> Generator[Order, None, None]:
        yield next(self.__iter__())

    def __len__(self) -> int:
        return len(self.orders)

    def _sort_orders_inplace(self) -> None:
        self.orders.sort(key=lambda order: order.timestamp)

    @property
    def _order_ids(self) -> list[str]:
        return [order.order_id for order in self.orders]

    def _get_order(self, order_id: str) -> Order:
        orders = [order for order in self.orders if order.order_id == order_id]
        assert len(orders) <= 1
        return orders[0]



from copy import deepcopy

import pandas as pd
from pytest_benchmark.fixture import BenchmarkFixture

from order_matching.matching_engine import MatchingEngine
from order_matching.order import LimitOrder, MarketOrder
from order_matching.orders import Orders
from order_matching.random import get_faker
from order_matching.side import Side
from order_matching.status import Status
from order_matching.trade import Trade


class TestMatchingEngine:
    def test_matching_with_no_orders(self) -> None:
        matching_engine = MatchingEngine()
        executed_trades = matching_engine.match(timestamp=pd.Timestamp.now())

        assert matching_engine.unprocessed_orders.bids == dict()
        assert matching_engine.unprocessed_orders.offers == dict()
        assert executed_trades.trades == []

    def test_matching_with_complete_order(self) -> None:
        order_book = MatchingEngine()

        assert order_book.unprocessed_orders.bids == dict()
        assert order_book.unprocessed_orders.offers == dict()
        assert order_book.unprocessed_orders.current_price == float("inf")

        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order_first = LimitOrder(
            side=Side.BUY, price=1.2, size=2.3, timestamp=timestamp, order_id="xyz", trader_id="x"
        )
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order_first])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {buy_order_first.price: Orders([buy_order_first])}
        assert order_book.unprocessed_orders.offers == dict()
        assert order_book.unprocessed_orders.current_price == float("inf")
        assert executed_trades.trades == []

        sell_order_first = LimitOrder(
            side=Side.SELL, price=3.4, size=5.6, timestamp=timestamp, order_id="abc", trader_id="x"
        )
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order_first])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {buy_order_first.price: Orders([buy_order_first])}
        assert order_book.unprocessed_orders.offers == {sell_order_first.price: Orders([sell_order_first])}
        assert order_book.unprocessed_orders.current_price == 2.3
        assert executed_trades.trades == []

        buy_order_second = LimitOrder(
            side=Side.BUY, price=buy_order_first.price, size=6.7, timestamp=timestamp, order_id="qwe", trader_id="x"
        )
        sell_order_second = LimitOrder(
            side=Side.SELL, price=5.9, size=9.3, timestamp=timestamp, order_id="abc", trader_id="x"
        )
        order_book.match(orders=deepcopy(Orders([buy_order_second])), timestamp=transaction_timestamp)
        order_book.match(orders=deepcopy(Orders([sell_order_second])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {
            buy_order_first.price: Orders([buy_order_first, buy_order_second])
        }
        assert order_book.unprocessed_orders.offers == {
            sell_order_first.price: Orders([sell_order_first]),
            sell_order_second.price: Orders([sell_order_second]),
        }
        assert order_book.unprocessed_orders.current_price == 2.3
        assert executed_trades.trades == []

    def test_matching_with_matching_offer_same_size(self) -> None:
        order_book = MatchingEngine(seed=42)
        size = 1
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        sell_order = LimitOrder(side=Side.SELL, price=3, size=size, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order.price: Orders([sell_order])}
        assert executed_trades.trades == []

        buy_order = LimitOrder(side=Side.BUY, price=4, size=size, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == [
            Trade(
                side=buy_order.side,
                size=size,
                price=sell_order.price,
                incoming_order_id=buy_order.order_id,
                book_order_id=sell_order.order_id,
                timestamp=transaction_timestamp,
                execution=buy_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]
        assert order_book.unprocessed_orders.current_price == float("inf")

    def test_matching_with_matching_bid_same_size(self) -> None:
        order_book = MatchingEngine(seed=42)
        size = 1
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order = LimitOrder(side=Side.BUY, price=4, size=size, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.offers == {}
        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order])}
        assert executed_trades.trades == []

        sell_order = LimitOrder(side=Side.SELL, price=3, size=size, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)
        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == [
            Trade(
                side=sell_order.side,
                size=size,
                price=buy_order.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]
        assert order_book.unprocessed_orders.current_price == float("inf")

    def test_matching_with_matching_offer_smaller_size(self) -> None:
        order_book = MatchingEngine(seed=42)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        sell_order = LimitOrder(side=Side.SELL, price=3, size=2, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order.price: Orders([sell_order])}
        assert executed_trades.trades == []

        buy_order = LimitOrder(side=Side.BUY, price=4, size=1, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)
        sell_order_modified = deepcopy(sell_order)
        sell_order_modified.size -= buy_order.size

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order.price: Orders([sell_order_modified])}
        assert executed_trades.trades == [
            Trade(
                side=buy_order.side,
                size=buy_order.size,
                price=sell_order.price,
                incoming_order_id=buy_order.order_id,
                book_order_id=sell_order.order_id,
                timestamp=transaction_timestamp,
                execution=buy_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]
        assert order_book.unprocessed_orders.current_price == 1.5

    def test_matching_with_matching_bid_smaller_size(self) -> None:
        order_book = MatchingEngine(seed=42)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order = LimitOrder(side=Side.BUY, price=4, size=2, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.offers == {}
        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order])}
        assert executed_trades.trades == []

        sell_order = LimitOrder(side=Side.SELL, price=3, size=1, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)
        buy_order_modified = deepcopy(buy_order)
        buy_order_modified.size -= sell_order.size

        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order_modified])}
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == [
            Trade(
                side=sell_order.side,
                size=sell_order.size,
                price=buy_order.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]
        assert order_book.unprocessed_orders.current_price == float("inf")

    def test_matching_with_matching_offer_bigger_size(self) -> None:
        order_book = MatchingEngine(seed=42)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        sell_order = LimitOrder(side=Side.SELL, price=3, size=1, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order.price: Orders([sell_order])}
        assert executed_trades.trades == []

        buy_order = LimitOrder(side=Side.BUY, price=4, size=2, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)
        buy_order_modified = deepcopy(buy_order)
        buy_order_modified.size -= sell_order.size

        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order_modified])}
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == [
            Trade(
                side=buy_order.side,
                size=buy_order_modified.size,
                price=sell_order.price,
                incoming_order_id=buy_order.order_id,
                book_order_id=sell_order.order_id,
                timestamp=transaction_timestamp,
                execution=buy_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]
        assert order_book.unprocessed_orders.current_price == float("inf")

    def test_matching_with_matching_offer_bigger_size_and_multiple_bids(self) -> None:
        order_book = MatchingEngine(seed=42)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        matching_order = LimitOrder(side=Side.BUY, price=5, size=1, timestamp=timestamp, order_id="c", trader_id="x")
        buy_orders = [
            LimitOrder(side=Side.BUY, price=3, size=2, timestamp=timestamp, order_id="a", trader_id="x"),
            LimitOrder(side=Side.BUY, price=4, size=3, timestamp=timestamp, order_id="b", trader_id="x"),
            matching_order,
        ]
        sell_order = LimitOrder(side=Side.SELL, price=5, size=10, timestamp=timestamp, order_id="d", trader_id="x")
        sell_orders = [sell_order]
        executed_trades = order_book.match(orders=deepcopy(Orders(buy_orders)), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {
            buy_orders[0].price: Orders([buy_orders[0]]),
            buy_orders[1].price: Orders([buy_orders[1]]),
            matching_order.price: Orders([matching_order]),
        }
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == []

        executed_trades = order_book.match(orders=deepcopy(Orders(sell_orders)), timestamp=transaction_timestamp)
        sell_order_modified = deepcopy(sell_order)
        sell_order_modified.size -= matching_order.size

        assert order_book.unprocessed_orders.bids == {
            buy_orders[0].price: Orders([buy_orders[0]]),
            buy_orders[1].price: Orders([buy_orders[1]]),
        }
        assert order_book.unprocessed_orders.offers == {sell_order.price: Orders([sell_order_modified])}
        assert executed_trades.trades == [
            Trade(
                side=sell_order.side,
                size=matching_order.size,
                price=sell_order.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=matching_order.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]

    def test_matching_with_matching_bid_bigger_size_and_multiple_offers(self) -> None:
        order_book = MatchingEngine(seed=42)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        matching_order = LimitOrder(side=Side.SELL, price=6, size=1, timestamp=timestamp, order_id="c", trader_id="x")
        sell_orders = [
            LimitOrder(side=Side.SELL, price=7, size=2, timestamp=timestamp, order_id="a", trader_id="x"),
            LimitOrder(side=Side.SELL, price=8, size=3, timestamp=timestamp, order_id="b", trader_id="x"),
            matching_order,
        ]
        buy_order = LimitOrder(side=Side.BUY, price=6, size=10, timestamp=timestamp, order_id="d", trader_id="x")
        buy_orders = [buy_order]
        executed_trades = order_book.match(orders=deepcopy(Orders(sell_orders)), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.offers == {
            sell_orders[0].price: Orders([sell_orders[0]]),
            sell_orders[1].price: Orders([sell_orders[1]]),
            matching_order.price: Orders([matching_order]),
        }
        assert order_book.unprocessed_orders.bids == {}
        assert executed_trades.trades == []

        executed_trades = order_book.match(orders=deepcopy(Orders(buy_orders)), timestamp=transaction_timestamp)
        buy_order_modified = deepcopy(buy_order)
        buy_order_modified.size -= matching_order.size

        assert order_book.unprocessed_orders.offers == {
            sell_orders[0].price: Orders([sell_orders[0]]),
            sell_orders[1].price: Orders([sell_orders[1]]),
        }
        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order_modified])}
        assert executed_trades.trades == [
            Trade(
                side=buy_order.side,
                size=matching_order.size,
                price=buy_order.price,
                incoming_order_id=buy_order.order_id,
                book_order_id=matching_order.order_id,
                timestamp=transaction_timestamp,
                execution=buy_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]

    def test_matching_with_matching_bid_bigger_size(self) -> None:
        order_book = MatchingEngine(seed=42)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order = LimitOrder(side=Side.BUY, price=4, size=1, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.offers == {}
        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order])}
        assert executed_trades.trades == []

        sell_order = LimitOrder(side=Side.SELL, price=3, size=2, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)
        sell_order_modified = deepcopy(sell_order)
        sell_order_modified.size -= buy_order.size

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order.price: Orders([sell_order_modified])}
        assert executed_trades.trades == [
            Trade(
                side=sell_order.side,
                size=sell_order_modified.size,
                price=buy_order.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]
        assert order_book.unprocessed_orders.current_price == 1.5

    def test_matching_with_multiple_matching_offers_bigger_size(self) -> None:
        order_book = MatchingEngine(seed=42)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        sell_order_first = LimitOrder(
            side=Side.SELL, price=3, size=1, timestamp=timestamp, order_id="abc", trader_id="x"
        )
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order_first])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order_first.price: Orders([sell_order_first])}
        assert executed_trades.trades == []

        sell_order_second = LimitOrder(
            side=Side.SELL, price=3, size=0.5, timestamp=timestamp, order_id="qwe", trader_id="x"
        )
        executed_trades = order_book.match(
            orders=deepcopy(Orders([sell_order_second])), timestamp=transaction_timestamp
        )

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {
            sell_order_first.price: Orders([sell_order_first, sell_order_second])
        }
        assert executed_trades.trades == []

        buy_order = LimitOrder(side=Side.BUY, price=4, size=2, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)
        buy_order_modified = deepcopy(buy_order)
        buy_order_modified.size -= sell_order_first.size + sell_order_second.size

        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order_modified])}
        assert order_book.unprocessed_orders.offers == {}
        faker = get_faker(seed=42)
        assert executed_trades.trades == [
            Trade(
                side=buy_order.side,
                size=sell_order_first.size,
                price=sell_order_first.price,
                incoming_order_id=buy_order.order_id,
                book_order_id=sell_order_first.order_id,
                timestamp=transaction_timestamp,
                execution=buy_order.execution,
                trade_id=faker.uuid4(),
            ),
            Trade(
                side=buy_order.side,
                size=sell_order_second.size,
                price=sell_order_second.price,
                incoming_order_id=buy_order.order_id,
                book_order_id=sell_order_second.order_id,
                timestamp=transaction_timestamp,
                execution=buy_order.execution,
                trade_id=faker.uuid4(),
            ),
        ]
        assert order_book.unprocessed_orders.current_price == float("inf")

    def test_matching_with_multiple_matching_bids_bigger_size(self) -> None:
        order_book = MatchingEngine(seed=66)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order_first = LimitOrder(side=Side.BUY, price=4, size=1, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order_first])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {buy_order_first.price: Orders([buy_order_first])}
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == []

        buy_order_second = LimitOrder(
            side=Side.BUY, price=4, size=0.5, timestamp=timestamp, order_id="qwe", trader_id="x"
        )
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order_second])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {
            buy_order_first.price: Orders([buy_order_first, buy_order_second])
        }
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == []

        sell_order = LimitOrder(side=Side.SELL, price=3, size=2, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)
        sell_order_modified = deepcopy(sell_order)
        sell_order_modified.size -= buy_order_first.size + buy_order_second.size

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order.price: Orders([sell_order_modified])}
        assert len(executed_trades.trades) == 2
        faker = get_faker(seed=66)
        assert executed_trades.trades == [
            Trade(
                side=sell_order.side,
                size=buy_order_first.size,
                price=buy_order_first.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order_first.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=faker.uuid4(),
            ),
            Trade(
                side=sell_order.side,
                size=buy_order_second.size,
                price=buy_order_second.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order_second.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=faker.uuid4(),
            ),
        ]

        assert order_book.unprocessed_orders.current_price == 1.5

    def test_matching_does_not_leave_zero_size_orders_behind(self) -> None:
        order_book = MatchingEngine(seed=66)
        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order_first = LimitOrder(side=Side.BUY, price=4, size=1, timestamp=timestamp, order_id="a", trader_id="x")
        buy_order_second = LimitOrder(side=Side.BUY, price=4, size=1, timestamp=timestamp, order_id="b", trader_id="x")
        executed_trades = order_book.match(
            orders=deepcopy(Orders([buy_order_first, buy_order_second])), timestamp=transaction_timestamp
        )

        assert order_book.unprocessed_orders.bids == {
            buy_order_first.price: Orders([buy_order_first, buy_order_second])
        }
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == []

        sell_order = LimitOrder(side=Side.SELL, price=3.5, size=1.5, timestamp=timestamp, order_id="qwe", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order])), timestamp=transaction_timestamp)
        sell_order_modified = deepcopy(sell_order)
        sell_order_modified.size -= buy_order_first.size
        buy_order_second_modified = deepcopy(buy_order_second)
        buy_order_second_modified.size -= sell_order.size - buy_order_first.size

        assert order_book.unprocessed_orders.bids == {buy_order_second.price: Orders([buy_order_second_modified])}
        assert order_book.unprocessed_orders.offers == {}
        faker = get_faker(seed=66)
        assert executed_trades.trades == [
            Trade(
                side=sell_order.side,
                size=buy_order_first.size,
                price=buy_order_first.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order_first.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=faker.uuid4(),
            ),
            Trade(
                side=sell_order.side,
                size=buy_order_first.size + buy_order_second.size - sell_order.size,
                price=buy_order_second.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order_second.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=faker.uuid4(),
            ),
        ]

    def test_matching_time_preference(self) -> None:
        order_book = MatchingEngine(seed=42)

        timestamp = pd.Timestamp(2022, 1, 5)
        timedelta = pd.Timedelta(1, unit="D")
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order_first = LimitOrder(side=Side.BUY, price=4, size=1, timestamp=timestamp, order_id="abc", trader_id="x")
        buy_order_second = LimitOrder(
            side=Side.BUY, price=4, size=1, timestamp=timestamp - timedelta, order_id="qwe", trader_id="x"
        )
        sell_order = LimitOrder(
            side=Side.SELL, price=4, size=0.5, timestamp=timestamp + timedelta, order_id="xyz", trader_id="x"
        )
        executed_trades = order_book.match(
            orders=deepcopy(Orders([sell_order, buy_order_first, buy_order_second])), timestamp=transaction_timestamp
        )
        buy_order_second_modified = deepcopy(buy_order_second)
        buy_order_second_modified.size -= sell_order.size

        assert buy_order_first.timestamp > buy_order_second.timestamp
        assert order_book.unprocessed_orders.bids == {
            buy_order_first.price: Orders([buy_order_second_modified, buy_order_first])
        }
        assert order_book.unprocessed_orders.offers == {}
        assert executed_trades.trades == [
            Trade(
                side=sell_order.side,
                size=sell_order.size,
                price=buy_order_second.price,
                incoming_order_id=sell_order.order_id,
                book_order_id=buy_order_second.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]

    def test_matching_with_market_orders_only(self) -> None:
        order_book = MatchingEngine(seed=42)

        assert order_book.unprocessed_orders.bids == dict()
        assert order_book.unprocessed_orders.offers == dict()
        assert order_book.unprocessed_orders.current_price == float("inf")

        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order_first = MarketOrder(side=Side.BUY, size=2.3, timestamp=timestamp, order_id="xyz", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order_first])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {buy_order_first.price: Orders([buy_order_first])}
        assert order_book.unprocessed_orders.offers == dict()
        assert order_book.unprocessed_orders.current_price == float("inf")
        assert executed_trades.trades == []

        sell_order_first = MarketOrder(side=Side.SELL, size=5.6, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order_first])), timestamp=transaction_timestamp)
        modified_sell_order_first = deepcopy(sell_order_first)
        modified_sell_order_first.size -= buy_order_first.size

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order_first.price: Orders([modified_sell_order_first])}
        assert order_book.unprocessed_orders.current_price == 0.0
        assert executed_trades.trades == [
            Trade(
                side=sell_order_first.side,
                size=buy_order_first.size,
                price=buy_order_first.price,
                incoming_order_id=sell_order_first.order_id,
                book_order_id=buy_order_first.order_id,
                timestamp=transaction_timestamp,
                execution=sell_order_first.execution,
                trade_id=get_faker(seed=42).uuid4(),
            )
        ]

    def test_matching_with_sell_market_order_after_limit_orders(self) -> None:
        order_book = MatchingEngine(seed=42)

        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_orders = [
            LimitOrder(side=Side.BUY, price=5.6, size=2.3, timestamp=timestamp, order_id="xyz", trader_id="x"),
            LimitOrder(side=Side.BUY, price=6.5, size=3.2, timestamp=timestamp, order_id="qwe", trader_id="x"),
        ]
        executed_trades = order_book.match(orders=deepcopy(Orders(buy_orders)), timestamp=transaction_timestamp)

        assert executed_trades.trades == []

        sell_order_first = MarketOrder(side=Side.SELL, size=10.0, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([sell_order_first])), timestamp=transaction_timestamp)
        modified_sell_order_first = deepcopy(sell_order_first)
        modified_sell_order_first.size -= buy_orders[0].size + buy_orders[1].size

        assert order_book.unprocessed_orders.bids == {}
        assert order_book.unprocessed_orders.offers == {sell_order_first.price: Orders([modified_sell_order_first])}
        assert order_book.unprocessed_orders.current_price == 0.0
        faker = get_faker(seed=42)
        assert executed_trades.trades == [
            Trade(
                side=sell_order_first.side,
                size=buy_orders[1].size,
                price=buy_orders[1].price,
                incoming_order_id=sell_order_first.order_id,
                book_order_id=buy_orders[1].order_id,
                timestamp=transaction_timestamp,
                execution=sell_order_first.execution,
                trade_id=faker.uuid4(),
            ),
            Trade(
                side=sell_order_first.side,
                size=buy_orders[0].size,
                price=buy_orders[0].price,
                incoming_order_id=sell_order_first.order_id,
                book_order_id=buy_orders[0].order_id,
                timestamp=transaction_timestamp,
                execution=sell_order_first.execution,
                trade_id=faker.uuid4(),
            ),
        ]

    def test_matching_with_buy_market_orders_after_limit_orders(self) -> None:
        order_book = MatchingEngine(seed=2)

        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        sell_orders = [
            LimitOrder(side=Side.SELL, price=5.6, size=2.3, timestamp=timestamp, order_id="xyz", trader_id="x"),
            LimitOrder(side=Side.SELL, price=6.5, size=3.2, timestamp=timestamp, order_id="qwe", trader_id="x"),
        ]
        executed_trades = order_book.match(orders=deepcopy(Orders(sell_orders)), timestamp=transaction_timestamp)

        assert executed_trades.trades == []

        buy_order_first = MarketOrder(side=Side.BUY, size=10.0, timestamp=timestamp, order_id="abc", trader_id="x")
        executed_trades = order_book.match(orders=deepcopy(Orders([buy_order_first])), timestamp=transaction_timestamp)
        modified_buy_order_first = deepcopy(buy_order_first)
        modified_buy_order_first.size -= sell_orders[0].size + sell_orders[1].size

        assert order_book.unprocessed_orders.bids == {buy_order_first.price: Orders([modified_buy_order_first])}
        assert order_book.unprocessed_orders.offers == {}
        assert order_book.unprocessed_orders.current_price == float("inf")
        faker = get_faker(seed=2)
        assert executed_trades.trades == [
            Trade(
                side=buy_order_first.side,
                size=sell_orders[0].size,
                price=sell_orders[0].price,
                incoming_order_id=buy_order_first.order_id,
                book_order_id=sell_orders[0].order_id,
                timestamp=transaction_timestamp,
                execution=buy_order_first.execution,
                trade_id=faker.uuid4(),
            ),
            Trade(
                side=buy_order_first.side,
                size=sell_orders[1].size,
                price=sell_orders[1].price,
                incoming_order_id=buy_order_first.order_id,
                book_order_id=sell_orders[1].order_id,
                timestamp=transaction_timestamp,
                execution=buy_order_first.execution,
                trade_id=faker.uuid4(),
            ),
        ]

    def test_matching_with_order_cancellation(self) -> None:
        order_book = MatchingEngine()

        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order = LimitOrder(side=Side.BUY, price=1.2, size=2.3, timestamp=timestamp, order_id="xyz", trader_id="x")
        order_book.match(orders=deepcopy(Orders([buy_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([buy_order])}

        cancel_buy_order = deepcopy(buy_order)
        cancel_buy_order.status = Status.CANCEL
        order_book.match(orders=deepcopy(Orders([cancel_buy_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {}

    def test_matching_with_order_cancellation_after_trading(self) -> None:
        order_book = MatchingEngine()

        timestamp = pd.Timestamp.now()
        transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
        buy_order = LimitOrder(side=Side.BUY, price=1.2, size=3.0, timestamp=timestamp, order_id="xyz", trader_id="x")
        sell_order = MarketOrder(side=Side.SELL, size=2.0, timestamp=timestamp, order_id="abc", trader_id="y")
        order_book.match(orders=deepcopy(Orders([buy_order, sell_order])), timestamp=transaction_timestamp)
        modified_buy_order = deepcopy(buy_order)
        modified_buy_order.size -= sell_order.size

        assert order_book.unprocessed_orders.bids == {buy_order.price: Orders([modified_buy_order])}

        cancel_buy_order = deepcopy(buy_order)

        assert cancel_buy_order != modified_buy_order
        assert cancel_buy_order.order_id == modified_buy_order.order_id
        assert cancel_buy_order.size != modified_buy_order.size

        cancel_buy_order.status = Status.CANCEL
        order_book.match(orders=deepcopy(Orders([cancel_buy_order])), timestamp=transaction_timestamp)

        assert order_book.unprocessed_orders.bids == {}

    def test_cancellation_of_expired_orders(self) -> None:
        matching_engine = MatchingEngine()

        timestamp = pd.Timestamp.now()
        timedelta = pd.Timedelta(1, unit="D")
        expiration = timestamp + timedelta
        order = LimitOrder(
            side=Side.BUY,
            price=1.2,
            size=3.0,
            timestamp=timestamp,
            expiration=expiration,
            order_id="xyz",
            trader_id="x",
        )
        matching_engine.match(orders=Orders([order]), timestamp=timestamp)

        assert matching_engine.unprocessed_orders.bids == {order.price: Orders([order])}
        assert matching_engine.unprocessed_orders.offers == dict()

        matching_engine.match(timestamp=timestamp + timedelta / 2)

        assert matching_engine.unprocessed_orders.bids == {order.price: Orders([order])}
        assert matching_engine.unprocessed_orders.offers == dict()

        matching_engine.match(timestamp=expiration)

        assert matching_engine.unprocessed_orders.bids == dict()
        assert matching_engine.unprocessed_orders.offers == dict()

    def test_matching_with_benchmark(self, random_orders: Orders, benchmark: BenchmarkFixture) -> None:
        order_book = MatchingEngine()
        benchmark(order_book.match, orders=random_orders, timestamp=random_orders.orders[-1].timestamp)



from copy import deepcopy

import pandas as pd

from order_matching.executed_trades import ExecutedTrades
from order_matching.execution import Execution
from order_matching.schemas import TradeDataSchema
from order_matching.side import Side
from order_matching.trade import Trade


class TestExecutedTrades:
    timestamp = pd.Timestamp.now()

    def test_init(self) -> None:
        executed_trades = ExecutedTrades()

        assert executed_trades.trades == list()

        trades = self._get_sample_trades()
        executed_trades = ExecutedTrades(trades=trades)

        assert executed_trades.trades == trades

    def test_add(self) -> None:
        executed_trades = ExecutedTrades()
        first_trade, second_trade = self._get_sample_trades()
        executed_trades.add(trades=[first_trade])

        assert executed_trades.trades == [first_trade]

        executed_trades.add(trades=[second_trade])

        assert executed_trades.trades == [first_trade, second_trade]

        executed_trades.add(trades=[second_trade, first_trade])

        assert executed_trades.trades == [first_trade, second_trade, second_trade, first_trade]

    def test_get(self) -> None:
        executed_trades = ExecutedTrades()
        first_trade, second_trade = self._get_sample_trades()
        executed_trades.add(trades=[first_trade, second_trade])

        assert executed_trades.get(timestamp=self.timestamp) == [first_trade, second_trade]

        third_trade = deepcopy(first_trade)
        third_trade.timestamp += pd.Timedelta(1, unit="D")
        executed_trades.add(trades=[third_trade])

        assert executed_trades.get(timestamp=self.timestamp) == [first_trade, second_trade]
        assert executed_trades.get(timestamp=third_trade.timestamp) == [third_trade]

    def test_to_frame(self) -> None:
        executed_trades = ExecutedTrades()

        pd.testing.assert_frame_equal(executed_trades.to_frame(), pd.DataFrame())

        first_trade, second_trade = self._get_sample_trades()
        executed_trades.add(trades=[first_trade, second_trade])

        TradeDataSchema.validate(executed_trades.to_frame(), lazy=True)

    def test_dunder_add(self) -> None:
        executed_trades_first = ExecutedTrades()
        first_trade, second_trade = self._get_sample_trades()
        executed_trades_first.add(trades=[first_trade])
        executed_trades_second = ExecutedTrades()
        executed_trades_second.add(trades=[second_trade])
        executed_trades_third = ExecutedTrades()
        executed_trades_third.add(trades=[first_trade])

        assert (executed_trades_first + executed_trades_second + executed_trades_third).trades == [
            first_trade,
            second_trade,
            first_trade,
        ]
        assert executed_trades_first.trades == [first_trade]
        assert executed_trades_second.trades == [second_trade]
        assert executed_trades_third.trades == [first_trade]

    def _get_sample_trades(self) -> list[Trade]:
        return [
            Trade(
                side=Side.SELL,
                price=0.5,
                size=0.9,
                timestamp=self.timestamp,
                incoming_order_id="a",
                book_order_id="x",
                execution=Execution.LIMIT,
                trade_id="t",
            ),
            Trade(
                side=Side.BUY,
                price=0.8,
                size=0.4,
                timestamp=self.timestamp,
                incoming_order_id="b",
                book_order_id="y",
                execution=Execution.LIMIT,
                trade_id="t",
            ),
        ]



from copy import deepcopy

import pandas as pd
import pytest

from order_matching.order import LimitOrder, Order
from order_matching.orders import Orders
from order_matching.schemas import OrderDataSchema
from order_matching.side import Side


class TestOrders:
    def test_init(self) -> None:
        order_queue = Orders()

        assert order_queue.orders == list()
        assert order_queue.is_empty

        orders = self._get_test_orders()
        order_queue = Orders(orders=orders)

        assert order_queue.orders != orders

        orders.sort(key=lambda order: order.timestamp)

        assert order_queue.orders == orders

    def test_dequeue(self) -> None:
        order_queue = Orders()
        orders = self._get_test_orders()
        order_queue.add(orders=orders)

        assert not order_queue.is_empty
        assert order_queue.orders == [orders[1], orders[0], orders[2], orders[3]]

        assert order_queue.dequeue() == orders[1]
        assert order_queue.dequeue() == orders[0]
        assert order_queue.dequeue() == orders[2]
        assert order_queue.dequeue() == orders[3]
        assert order_queue.is_empty

    @pytest.mark.parametrize("ids_to_remove", [[0], [2], [1, -1]])
    def test_remove(self, ids_to_remove: list[int]) -> None:
        orders = self._get_test_orders()
        orders_to_remove = deepcopy([orders[i] for i in ids_to_remove])
        order_queue = Orders()
        order_queue.remove(orders=orders_to_remove)

        assert order_queue == Orders()

        order_queue = Orders(orders=orders)
        for order in orders_to_remove:
            order.size /= 2
        order_queue.remove(orders=orders_to_remove)
        order_ids_to_remove = [order.order_id for order in orders_to_remove]
        expected_leftover_orders = [order for order in orders if order.order_id not in order_ids_to_remove]
        expected_queue = Orders(orders=expected_leftover_orders)

        assert order_queue == expected_queue

    def test_to_frame(self) -> None:
        order_queue = Orders()
        orders = self._get_test_orders()
        order_queue.add(orders=orders)

        OrderDataSchema.validate(order_queue.to_frame(), lazy=True)

    def test_dunder_add_and_len(self) -> None:
        order_queue_first = Orders()

        assert len(order_queue_first) == 0

        first_order, second_order = self._get_test_orders()[:2]
        order_queue_first.add(orders=[first_order])
        order_queue_second = Orders()
        order_queue_second.add(orders=[second_order])
        order_queue_third = Orders()
        order_queue_third.add(orders=[first_order])
        order_queue_all = Orders()
        order_queue_all.add(orders=[first_order, second_order, second_order])

        assert len(order_queue_first) == 1
        assert len(order_queue_second) == 1
        assert len(order_queue_third) == 1
        assert len(order_queue_all) == 3
        assert (order_queue_first + order_queue_second + order_queue_third).orders == [
            second_order,
            first_order,
            first_order,
        ]
        assert order_queue_first.orders == [first_order]
        assert order_queue_second.orders == [second_order]
        assert order_queue_third.orders == [first_order]

    @staticmethod
    def _get_test_orders() -> list[Order]:
        timestamp = pd.Timestamp.now()
        return [
            LimitOrder(side=Side.SELL, price=4.0, size=10.0, timestamp=timestamp, order_id="a", trader_id="x"),
            LimitOrder(
                side=Side.BUY,
                price=4.0,
                size=12.0,
                timestamp=timestamp - pd.Timedelta(1, unit="D"),
                order_id="b",
                trader_id="x",
            ),
            LimitOrder(side=Side.BUY, price=4.0, size=7.0, timestamp=timestamp, order_id="c", trader_id="x"),
            LimitOrder(side=Side.SELL, price=3.0, size=8.0, timestamp=timestamp, order_id="d", trader_id="x"),
        ]



"""Generate the code reference pages.

Source: https://mkdocstrings.github.io/recipes/#automatic-code-reference-pages
"""

from pathlib import Path

import mkdocs_gen_files

SRC = "src"

nav = mkdocs_gen_files.Nav()

for path in sorted(Path(SRC).rglob("*.py")):
    module_path = path.relative_to(SRC).with_suffix("")
    doc_path = path.relative_to(SRC).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        continue
    if parts[-1] == "__main__":
        continue

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, mode="w") as fd:
        print("::: " + ".".join(parts), file=fd)

    mkdocs_gen_files.set_edit_path(name=full_doc_path, edit_name=path)

with mkdocs_gen_files.open("reference/SUMMARY.md", mode="w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



import pandas as pd

from order_matching.executed_trades import ExecutedTrades
from order_matching.order import Order
from order_matching.order_book import OrderBook
from order_matching.orders import Orders
from order_matching.random import get_faker
from order_matching.status import Status
from order_matching.trade import Trade


class MatchingEngine:
    """Order Book Matching Engine.

    Parameters
    ----------
    seed
        Random seed

    Examples
    --------
    >>> from pprint import pp
    >>> from order_matching.matching_engine import MatchingEngine
    >>> from order_matching.order import LimitOrder
    >>> from order_matching.side import Side
    >>> matching_engine = MatchingEngine(seed=123)
    >>> timestamp = pd.Timestamp("2023-01-01")
    >>> transaction_timestamp = timestamp + pd.Timedelta(1, unit="D")
    >>> buy_order = LimitOrder(side=Side.BUY, price=1.2, size=2.3, timestamp=timestamp, order_id="a", trader_id="x")
    >>> sell_order = LimitOrder(side=Side.SELL, price=0.8, size=1.6, timestamp=timestamp, order_id="b", trader_id="y")
    >>> executed_trades = matching_engine.match(orders=Orders([buy_order, sell_order]), timestamp=transaction_timestamp)
    >>> pp(executed_trades.trades)
    [Trade(side=SELL,
           price=1.2,
           size=1.6,
           incoming_order_id='b',
           book_order_id='a',
           execution=LIMIT,
           trade_id='c4da537c-1651-4dae-8486-7db30d67b366',
           timestamp=Timestamp('2023-01-02 00:00:00'))]
    """

    def __init__(self, seed: int = None) -> None:
        self._seed = seed
        self._faker = get_faker(seed=seed)
        self._queue = Orders()
        self.unprocessed_orders = OrderBook()
        self._timestamp: pd.Timestamp | None = None

    def match(self, timestamp: pd.Timestamp, orders: Orders = None) -> ExecutedTrades:
        """Match incoming orders in price-time priority.

        Parameters
        ----------
        timestamp
            Timestamp of order matching
        orders
            Incoming orders. Will be matched with existing ones on the order book in

        Returns
        -------
        ExecutedTrades
            Executed trades storage object
        """
        self._timestamp = timestamp
        self._queue += orders if orders else Orders()
        self._queue += self._get_expired_orders()
        trades = ExecutedTrades()
        while not self._queue.is_empty:
            trades += self._match(order=self._queue.dequeue())
        return trades

    def _get_expired_orders(self) -> Orders:
        orders: list[Order] = list()
        for timestamp in filter(lambda t: t <= self._timestamp, self.unprocessed_orders.orders_by_expiration.keys()):
            orders.extend(self.unprocessed_orders.orders_by_expiration[timestamp])
        for order in orders:
            order.status = Status.CANCEL
        return Orders(orders)

    def _match(self, order: Order) -> ExecutedTrades:
        if order.status == Status.CANCEL:
            self.unprocessed_orders.remove(incoming_order=order)
            return ExecutedTrades()
        elif self.unprocessed_orders.matching_order_exists(incoming_order=order):
            return self._execute_trades(incoming_order=order)
        else:
            self.unprocessed_orders.append(incoming_order=order)
            return ExecutedTrades()

    def _execute_trades(self, incoming_order: Order) -> ExecutedTrades:
        trades = ExecutedTrades()
        for price in self.unprocessed_orders.get_matching_sorted_opposite_side_prices(incoming_order=incoming_order):
            trades += self._execute_trades_for_one_price(incoming_order=incoming_order, price=price)
        if incoming_order.size > 0:
            self.unprocessed_orders.append(incoming_order=incoming_order)
        return trades

    def _execute_trades_for_one_price(self, incoming_order: Order, price: float) -> ExecutedTrades:
        opposite_side_orders = self.unprocessed_orders.get_opposite_side_orders(incoming_order=incoming_order)
        trades, zero_size_orders = list(), list()
        for book_order in opposite_side_orders[price]:
            if incoming_order.size > 0:
                trades.append(self._execute_trade(incoming_order=incoming_order, book_order=book_order))
            if book_order.size == 0:
                zero_size_orders.append(book_order)
        opposite_side_orders[price].remove(orders=zero_size_orders)
        if len(list(filter(lambda order: order.size > 0, opposite_side_orders[price]))) == 0:
            opposite_side_orders.pop(price)
        return ExecutedTrades(trades=trades)

    def _execute_trade(self, incoming_order: Order, book_order: Order) -> Trade:
        trade = Trade(
            side=incoming_order.side,
            price=book_order.price,
            size=min(incoming_order.size, book_order.size),
            incoming_order_id=incoming_order.order_id,
            book_order_id=book_order.order_id,
            timestamp=self._timestamp,
            execution=incoming_order.execution,
            trade_id=self._faker.uuid4(),
        )
        incoming_order.size = max(0.0, incoming_order.size - trade.size)
        book_order.size = max(0.0, book_order.size - trade.size)
        return trade
